/*
 * connection.cpp
 *
 * Copyright 2001 Matthew Stephenson
 * 
 * This file is part of Zic.
 * 
 * Zic is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * Zic is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Zic; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Look, feel, and userlist icons shamelessly stolen from X-Chat 
 * (www.xchat.org)
 * 
 */

#include "connection.h"

connection::connection()
{
  sock = new socket();
  connect(sock, SIGNAL(readyLine()), this, SLOT(readSocket()));
  connect(sock, SIGNAL(connected()), this, SLOT(login()));
  connect(sock, SIGNAL(connectionClosed()), this, SLOT(socketDisconnect()));

  serverNick = "";
  me = new ircuser();
  b_connected = false;
  b_loggedIn = false;

}

void connection::connectToServer( QString server, int port, QString nick, 
			  QString username, QString realname)
{
  if(nick.length() == 0) {
    emit errormsg("Please specify a nickname with /nick");
    return;
  }
  me->set_nick(nick);

  if(username.length() == 0)
    username = "zic"; //FIXME I'm sure I shouldn't be hardcoding this
  me->set_username(username);

  if(realname.length() == 0)
    realname = "Zaurus Internet Chat";
  me->set_realname(realname);

  if(server.length() == 0) {
    emit errormsg("Please specify a server");
    return;
  }
  this->serverIP = server;

  if(port == 0)
    port = 6667;
  this->serverPort = port;

  //FIXME should be sent to the server tab
  emit debug("Connecting to Server...");
  sock->connectToHost( server, port );

}

void connection::login() 
{
  //FIXME should be sent to the server tab
  emit debug("Connected, Logging in...");
  //FIXME should be the real hostname
  sock->sendLine("USER " + me->username() + " " + 
		 "HOSTNAME" + " " + serverIP + " :" + me->realname());
  sock->sendLine("NICK " + me->nick());

  if(autojoin.count() > 0)
    sock->sendLine("JOIN " + autojoin.join(","));

  b_connected = true;
  
}


void connection::readSocket()
{
  QString line;
  
  while(sock->canReadLine()) {
    line = sock->readLine();
    emit(gotRawLine(line));
    parseLine(line);
  }
}

void connection::parseLine( QString line )
{

  ircmessage *msg = new ircmessage( line );
  
  // Bigass switch statement:
  
  if(msg->command() == "001") {
    // use the 001 to work out what the server is, and to flag that we're logged in
    serverNick = msg->from()->nick();
    b_loggedIn = true;
    emit message( msg );
  }
  else if(msg->command() == "PING") {
    emit debug("PING? PONG!");
    sock->sendLine("PONG " + msg->param(0));
  }
  else if(msg->command() == "PRIVMSG") {
    if(msg->is_ctcp())      
      emit ctcp( msg );
    else 
      emit privmsg( msg );   
  }
  else if(msg->command() == "NOTICE") {
    if(msg->is_ctcp())      
      emit ctcpnotice( msg );
    else 
      emit notice( msg );
  }
  else if(msg->command() == "JOIN") {
    emit join( msg );
  }
  else if(msg->command() == "PART") {
    emit part( msg );
  }
  else if(msg->command() == "QUIT") {
    emit quit( msg );
  }
  else if(msg->command() == "353") { //RPL_NAMREPLY
    emitNames(msg);
  }
  else if(msg->command() == "366") { //RPL_ENDOFNAMES
    emit rpl_endofnames(msg->param(1));
  }
  else if(msg->command() == "MODE") {
    parseModes( msg );
  }
  else if(msg->command() == "NICK") {
    parseNick( msg );
  }
  else if(msg->command() == "KICK") {
    emit kick( msg );
  }
  else {
    emit message( msg );
  }

  delete msg;
}


void connection::socketDisconnect()
{
  b_connected = false;
  b_loggedIn = false;
  serverNick = "";
  emit disconnected();
}


void connection::parseCommand( QString line, QString fromChannel = "" )
{
  
  QString command;

  // This should only get passed lines with a "/" on the front
  if(line.left(1) == "/") {
    line = line.right(line.length() - 1);
  }

  QTextIStream stream(&line);

  stream >> command;
  
  command = command.upper();

  emit debug("Got command: " + command );

  // bigass switch statement coming right up :)

  if(command == "SERVER") {
    QString my_server;
    int my_port = 0;
    // FIXME parameter error checking goes here
    stream >> my_server;
    stream >> my_port;
    if(my_port == 0)
      my_port = DEFAULT_PORT;
    connectToServer( my_server, my_port, me->nick(), me->username(), me->realname());
  }
  else if(command == "QUIT") {
    stream.skipWhiteSpace();
    sock->sendLine("QUIT :" + stream.readLine());
  } 
  else if(command == "QUOTE") {
    stream.skipWhiteSpace();
    sock->sendLine(stream.readLine());
  }
  else if(command == "MSG") {
    QString to;
    stream >> to;
    stream.skipWhiteSpace();
    sock->sendLine("PRIVMSG " + to + " :" + stream.readLine());
  }
  else if(command == "JOIN") {
    stream.skipWhiteSpace();
    sock->sendLine("JOIN :" + stream.readLine());
  }    
  else if(command == "PART") {
    if(!(isChannel(fromChannel)))
      return;
    QString channel, reason;
    stream >> channel;
    if(channel != "") {
      stream.skipWhiteSpace();
      reason = stream.readLine();
      sock->sendLine("PART " + channel + " :" + reason);
    }
    else {
      sock->sendLine("PART " + fromChannel + " :" + reason);
    }
  }
  else if(command == "MODE") {
    QString line;
    stream.skipWhiteSpace();
    line = stream.readLine();
    sock->sendLine("MODE " + line);
  }
  else if(command == "NICK") {
    QString line;
    stream.skipWhiteSpace();
    line = stream.readLine();
    if(b_connected) {
      sock->sendLine("NICK " + line);
    }
    else {
      me->set_nick(line);
    }
  }
  else if(command == "KICK") {
    if(!(isChannel(fromChannel)))
      return;
    QString nick, reason;
    stream >> nick;
    stream.skipWhiteSpace();
    reason = stream.readLine();
    emit debug("KICK " + fromChannel + " " + nick + " :" + reason);
    sock->sendLine("KICK " + fromChannel + " " + nick + " :" + reason);
  }
  else if(command == "CTCP") {
    QString to;
    stream >> to;
    stream.skipWhiteSpace();
    sendCtcp(to, stream.readLine());
  }
  else if(command == "ME") {
    stream.skipWhiteSpace();
    sendCtcp(fromChannel, "ACTION " + stream.readLine());
  }

}


void connection::emitNames( ircmessage *msg )
{
  QStringList *names = new QStringList;
  for(int i=3; i < msg->paramcount(); i++) {
    names->append(msg->param(i, true));
    emit debug("name: " + msg->param(i, true));
  }
  emit rpl_namreply( msg->param(2), names );
  delete names;
}



QString connection::getServerNick()
{
  return serverNick;
}

bool connection::loggedIn()
{
  return b_loggedIn;
}

void connection::sendPrivMsg( QString to, QString message )
{
  sock->sendLine("PRIVMSG " + to + " :" + message );
}

void connection::sendCtcp( QString to, QString message )
{
  sock->sendLine("PRIVMSG " + to + " :" + QChar(1) + message + QChar(1));
}

void connection::sendCtcpNotice( QString to, QString message )
{
  sock->sendLine("NOTICE " + to + " :" + QChar(1) + message + QChar(1));
}


void connection::parseModes( ircmessage *msg )
{
  ircmode * m_modes;
  ircmode * m_tmp;
  QChar modechar;
  QChar setchar = '+';
  QString arg;
  unsigned int i=0, j=0;
  
  m_modes = new ircmode( msg );
  emit modes(m_modes);
  

  emit debug("Got modes: " + m_modes->modes()
	     + " " + m_modes->args() + " on " 
	     + m_modes->target() + " by " 
  	     + m_modes->from()->nick());


   
  if(isChannel(m_modes->target())) {
    // Channel modes
    for(i=0; i < m_modes->modes().length(); i++) {
      modechar = m_modes->modes().at(i).latin1();
      
      switch(modechar) {
      case '+':
      case '-':
	setchar = modechar;
	break;
      case 'o':
      case 'v':
      case 'h':
      case 'k':
      case 'l':
	arg = m_modes->arg(j);
	j++;

	  m_tmp = new ircmode(&ircmessage(":" + msg->from()->nick() + " MODE " +
					 m_modes->target() + " " +
					 setchar + modechar + " " +
					 arg));
	  emit mode( m_tmp );
	  delete m_tmp;

	  emit debug(QString("Got mode: +") + QString(modechar) + 
		     " " + arg + " on " + m_modes->target() + 
		     " by " + m_modes->from()->nick());
	break;
      default:

	m_tmp = new ircmode(&ircmessage(":" + msg->from()->nick() + " MODE " +
					m_modes->target() + " " +
					setchar + modechar + " "));
	emit mode( m_tmp );
	delete m_tmp;
	
	emit debug(QString("Got mode: +") + QString(modechar) + 
		   " on " + m_modes->target() + 
		   " by " + m_modes->from()->nick());
	break;
      }
      
    }
  }
  else {
    for(i=0; i < m_modes->modes().length(); i++) {
      modechar = m_modes->modes().at(i).latin1();
      
      switch(modechar) {
      case '+':
      case '-':
	setchar = modechar;
	break;
      default:
	m_tmp = new ircmode(&ircmessage(":" + msg->from()->nick() + " MODE " +
					m_modes->target() + " " +
					setchar + modechar + " "));
	emit mode( m_tmp );
	delete m_tmp;
	
	emit debug(QString("Got mode: +") + QString(modechar) + 
		   " on " + m_modes->target() + 
		   " by " + m_modes->from()->nick());
	break;
      }
    }
  }
  
  delete m_modes;

}

bool connection::isChannel(const QString name)
{
  QString chars = QString(CHANNEL_CHARS);
  uint i;
  
  for(i = 0; i < chars.length(); i++) {
    if(name.at(0) == chars.at(i)) {
      return true;
    }
  }
  return false;
}


void connection::parseNick( ircmessage *msg )
{
  // is it us changing nick?
  if(msg->from()->nick() == me->nick()) {
    me->set_nick(msg->param(0, true));
  }
  emit nick( msg );
  emit debug("Got nick change - " + msg->from()->nick() + " -> " 
	     + msg->param(0, true));
}
