/*
 * ircsession.cpp
 *
 * Copyright 2001 Matthew Stephenson
 * 
 * This file is part of Zic.
 * 
 * Zic is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * Zic is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Zic; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Look, feel, and userlist icons shamelessly stolen from X-Chat 
 * (www.xchat.org)
 * 
 */

#include <qvbox.h>
#include "ircsession.h"
#include "plaintab.h"
#include "connectbox.h"

ircsession::ircsession( zic *parent) 
{
  this->parent = parent;
  servertab = addNewPlainTab( "Server" ); 

  conn = new connection();
  connect(conn, SIGNAL(disconnected()), this, SLOT(socketDisconnect()));
  connect(conn, SIGNAL(errormsg(QString)), this, SLOT(errormsg(QString)));
  connect(conn, SIGNAL(privmsg(ircmessage*)), this, SLOT(parsePrivmsg(ircmessage*)));
  connect(conn, SIGNAL(notice(ircmessage*)), this, SLOT(parseNotice(ircmessage*)));
  //FIXME do we need to parse ctcp notices differently?
  connect(conn, SIGNAL(ctcpnotice(ircmessage*)), this, SLOT(parseNotice(ircmessage*)));
  connect(conn, SIGNAL(join(ircmessage*)), this, SLOT(parseJoin(ircmessage*)));
  connect(conn, SIGNAL(part(ircmessage*)), this, SLOT(parsePart(ircmessage*)));
  connect(conn, SIGNAL(quit(ircmessage*)), this, SLOT(parseQuit(ircmessage*)));
  connect(conn, SIGNAL(kick(ircmessage*)), this, SLOT(parseKick(ircmessage*)));
  connect(conn, SIGNAL(ctcp(ircmessage*)), this, SLOT(parseCtcp(ircmessage*)));
  connect(conn, SIGNAL(mode(ircmode *)), this, SLOT(parseMode(ircmode *)));
  connect(conn, SIGNAL(message(ircmessage*)), this, SLOT(parseMessage(ircmessage*)));
  connect(conn, SIGNAL(rpl_namreply(QString, QStringList*)), 
	  this, SLOT(rpl_namreply(QString, QStringList*)));
  connect(conn, SIGNAL(rpl_endofnames(QString)), this, SLOT(rpl_endofnames(QString)));
  connect(conn, SIGNAL(nick(ircmessage *)), this, SLOT(parseNick(ircmessage *)));

  
#ifdef IRCDEBUG
  rawtab = addNewPlainTab( "Raw" ); 
  connect( conn, SIGNAL(gotRawLine(QString)), rawtab, SLOT(addRawLine(QString)));

  debugtab = addNewPlainTab( "Debug" ); 
  connect( conn, SIGNAL(debug(QString)), debugtab, SLOT(addRawLine(QString)));
#endif

  // FIXME test code
  //conn->me->set_nick("foo");
  //conn->me->set_username("foo");
  //conn->me->set_realname("test client");

 
}

plaintab * ircsession::addNewPlainTab( QString label )
{
  plaintab *tab = new plaintab( parent, label);
  
  tabs.append( tab );
    connect(tab, SIGNAL(gotUserInput(QString, irctab*)), 
    	  this, SLOT(userInput(QString, irctab*)));
  connect(tab, SIGNAL(amTopTab(irctab*)), this, SLOT(newTopTab(irctab*)));
  return tab;
}


channeltab * ircsession::addNewChannelTab( QString label )
{
  ircchannel *channel = new ircchannel( label );
  channels.append( channel );

  channeltab *tab = new channeltab( parent, channel, label );
  tabs.append( tab );

    connect(tab, SIGNAL(gotUserInput(QString, irctab*)), 
    	  this, SLOT(userInput(QString, irctab*)));
    connect(conn, SIGNAL(modes(ircmode *)), tab, SLOT(modes(ircmode *)));
    connect(conn, SIGNAL(mode(ircmode *)), tab, SLOT(mode(ircmode *)));
  connect(tab, SIGNAL(amTopTab(irctab*)), this, SLOT(newTopTab(irctab*)));
  return tab;
}


void ircsession::removeTab()
{
  tabs.remove(tabtoremove);
  delete tabtoremove;
  
}

void ircsession::removeTab( irctab *tab )
{
#ifdef IRCDEBUG
  //debugtab->addRawLine("Removing tab " + tab->name());
#endif
  // just saying "delete tab;" will obviously cause a segfault.
  // because this gets called from an emit, from the tab itself
  // So this is a nasty hack with a 1 millisecond timer :)
  tabtoremove = tab;
  QTimer::singleShot( 1, this, SLOT(removeTab()) );
}


void ircsession::userInput( QString text, irctab* from )
{
#ifdef IRCDEBUG
  debugtab->addRawLine("Got User Input: " + text +" from " + from->label());
#endif

  // If it has a / on the front... 
  if(text.left(1) == "/") 
    if(text.left(6).upper() == "/CLOSE" && (from != servertab)) {
#ifdef IRCDEBUG
      debugtab->addRawLine("Removing tab " + from->label());
#endif
      removeTab(from);
    }
    else if(text.left(4).upper() == "/ME " && (from != servertab)) {
      from->action(conn->me->nick(), text.right(text.length() - 4));
      conn->parseCommand(text, from->label());
    }
    else //pass it on to the connection object
      conn->parseCommand(text, from->label());
  else {
    // doesn't have a / on the front, so send a privmsg
    if(text.length() == 0)
      return;
    conn->sendPrivMsg(from->label(), text);
    from->privmsg(conn->me->nick(), text);
  }

}

void ircsession::socketDisconnect()
{
  // FIXME this should broadcast
  for(irctab *tab=tabs.first(); tab != 0; tab=tabs.next()) {
    tab->message("Disconnected");
  }
}

void ircsession::newTopTab( irctab *tab )
{
  toptab = tab;
}



void ircsession::parsePrivmsg( ircmessage *msg )
{

  // see if it's to us...
  if(msg->param(0) == conn->me->nick()) {
    // okay, it's to us - do we have a tab for this person?
    if(irctab *tab = findTab(msg->from()->nick())) {
      tab->privmsg(msg->from()->nick(), msg->params(1, true));
      return;
    }
    // so we don't have a tab for this person - create one
#ifdef IRCDEBUG
    debugtab->addRawLine("No tab for " + msg->from()->nick() + ", creating...");
#endif
    plaintab *tab = addNewPlainTab(msg->from()->nick());
    tab->privmsg(msg->from()->nick(), msg->params(1, true));
  }
  else {
    // it's not to us, so it must be to a channel we're in. Do we have a tab?
    if(irctab *tab = findTab(msg->param(0,true))) {
      tab->privmsg(msg->from()->nick(), msg->params(1, true));
      return;
    }
    // so we don't have a channel tab. Odd. Create one.
#ifdef IRCDEBUG
    debugtab->addRawLine("No tab for channel " + msg->param(0, true) + ", creating...");
#endif      
    channeltab *tab = addNewChannelTab(msg->param(0, true));
    tab->privmsg(msg->from()->nick(), msg->params(1, true));
  }
  
}


void ircsession::parseNotice( ircmessage *msg )
{

  // see if it's from the server
  if((conn->loggedIn() == false) || (msg->from()->nick() == conn->getServerNick())) {
    servertab->notice(msg->from()->nick(), msg->params(1, true));
    return;
  }

  // see if it's to us...
  if(msg->param(0) == conn->me->nick()) {
    // do we have a tab for this person
    if(irctab *tab = findTab(msg->from()->nick())) {
      tab->notice(msg->from()->nick(), msg->params(1, true));
      return;
    }
    // so we don't have a tab for this person - print to the top tab
    toptab->notice(msg->from()->nick(), msg->params(1, true));
  }
  else {
    // it's not to us, so it must be to a channel we're in. Do we have a tab?
    if(irctab *tab = findTab(msg->param(0,true))) {
      static_cast<plaintab *>(tab)->notice(msg->from()->nick(), msg->params(1, true));
      return;
    }
    // so we don't have a channel tab. Odd. Create one.
#ifdef IRCDEBUG
    debugtab->addRawLine("No tab for channel " + msg->param(0, true) + ", creating...");
#endif
    channeltab *tab = addNewChannelTab(msg->param(0, true));
    tab->notice(msg->from()->nick(), msg->params(1, true));
  }
}


void ircsession::parseJoin( ircmessage *msg )
{
  // is it us joining something?
  
  if(msg->from()->nick() == conn->me->nick()) {
    channeltab *tab = addNewChannelTab(msg->param(0, true));
    // tab->join(msg->from()); 
    //FIXME could show us joining, but it results in multiple userlist entries :/
  }
  else {
    channeltab* cpt;
    // well, I WOULD use dynamic_cast, but qt seems to need -fno-rtti
    cpt = static_cast<channeltab *>(findTab(msg->param(0, true)));
    if(cpt)
      cpt->join(msg->from());
    
  }
}

void ircsession::parsePart( ircmessage *msg )
{
  // is it us parting something?
  
  if(msg->from()->nick() == conn->me->nick()) {
    irctab *tab = findTab(msg->param(0, true));
    removeTab(tab);
  }
  else {
    // well, I WOULD use dynamic_cast, but qt seems to need -fno-rtti
    channeltab* cpt = static_cast<channeltab *>(findTab(msg->param(0, true)));
    if(cpt)
      cpt->part(msg->from(), msg->params(1, true));
  }
}

void ircsession::parseQuit( ircmessage *msg )
{
  // is it us quitting? 
  if(msg->from()->nick() == conn->me->nick()) {
    //FIXME eh, what the **** do we do here?
  }
  else {
    channeltab *tab;
    for(ircchannel *p_channel=channels.first(); p_channel != 0; 
	p_channel=channels.next()) {
      if(p_channel->findUser(msg->from()->nick())) {
	// well, I WOULD use dynamic_cast, but qt seems to need -fno-rtti      
	tab = static_cast<channeltab *>(findTab(p_channel->name()));
	tab->quit(msg->from(), msg->params(true));
      }
    }
    
  }
}


void ircsession::parseKick( ircmessage *msg ) {
  //FIXME RFC allows for multiple kicks in one go
  //FIXME need to do stuff for when WE get kicked
  irctab *tab = findTab( msg->param(0, true) );
  if(tab) {
    // well, I WOULD use dynamic_cast, but qt seems to need -fno-rtti
    channeltab *cpt = static_cast<channeltab *>(tab);
    cpt->kick(msg->from(), msg->param(1), msg->params(2, true));
  }  
}


//  All unloved messages end up here :)
void ircsession::parseMessage( ircmessage *msg )
{
  //FIXME this should look for the colon, and display everything after it
  servertab->message( msg->params(1, true) );
}

void ircsession::rpl_namreply( QString channel, QStringList *names )
{
  irctab *tab = findTab( channel );
  if(tab) {
    // well, I WOULD use dynamic_cast, but qt seems to need -fno-rtti
    channeltab *cpt = static_cast<channeltab *>(tab);
    cpt->names(names);
  }
}

void ircsession::rpl_endofnames( QString channel )
{
  irctab *tab = findTab( channel );
  if(tab) {
    // well, I WOULD use dynamic_cast, but qt seems to need -fno-rtti
    channeltab *cpt = static_cast<channeltab *>(tab);    
    cpt->endofnames();
  }
}

irctab * ircsession::findTab( QString name )
{
  for(irctab *tab=tabs.first(); tab != 0; tab=tabs.next()) {
    if(tab->label().lower() == name.lower()) {
      return tab;
    }
  }
  return 0;
}

ircchannel * ircsession::findChannel( QString name )
{
  for(ircchannel *channel=channels.first(); channel != 0; 
      channel=channels.next()) {
    if(channel->name().lower() == name.lower()) {
      return channel;
    }
  }
  return 0;
}

void ircsession::errormsg( QString msg )
{
  toptab->message(msg);
}

void ircsession::parseNick( ircmessage *msg )
{

  //FIXME this needs to deal with nicks in query windows too

  channeltab *tab;
  for(ircchannel *p_channel=channels.first(); p_channel != 0; 
      p_channel=channels.next()) {
    if(p_channel->findUser(msg->from()->nick())) {
      // well, I WOULD use dynamic_cast, but qt seems to need -fno-rtti      
      tab = static_cast<channeltab *>(findTab(p_channel->name()));
      tab->nick(msg->from()->nick(), msg->param(0, true));
    }
  }
}

void ircsession::showConnectBox()
{
  QString temp_nick, temp_server;
  uint temp_port;
  QStringList temp_channels;
  connectbox foo(&temp_nick, &temp_server, &temp_port, &temp_channels);
  foo.exec();
  
  if(foo.result() == QDialog::Accepted) {
#ifdef IRCDEBUG
    debugtab->addRawLine("Server = " + temp_server);
#endif
    if((temp_nick.length() > 0) && (temp_server.length() > 0)) {
      conn->autojoin = temp_channels;
      conn->connectToServer( temp_server, temp_port, temp_nick,
			     conn->me->username(), conn->me->realname());
    }
  }
}


void ircsession::parseCtcp( ircmessage *msg )
{
#ifdef IRCDEBUG
  debugtab->addRawLine("Got CTCP " + msg->params(true));
#endif
  if(msg->param(1, true).lower() == "version") {
    servertab->message("Recieved CTCP VERSION from " + msg->from()->nick());
    conn->sendCtcpNotice(msg->from()->nick(), QString("VERSION Zic ") + ZICVERSION);
  }
  else if(msg->param(1, true).lower() == "action") {
    parseAction( msg );
  }
}


void ircsession::parseAction( ircmessage *msg )
{
  // see if it's to us...
  if(msg->param(0) == conn->me->nick()) {
    // do we have a tab for this person
    if(irctab *tab = findTab(msg->from()->nick())) {
      tab->action(msg->from()->nick(), msg->params(2, true));
      return;
    }
    // so we don't have a tab for this person - create one
#ifdef IRCDEBUG
    debugtab->addRawLine("No tab for " + msg->from()->nick() + ", creating...");
#endif
    plaintab *tab = addNewPlainTab(msg->from()->nick());
    tab->action(msg->from()->nick(), msg->params(2, true));
  }
  else {
    // it's not to us, so it must be to a channel we're in. Do we have a tab?
    if(irctab *tab = findTab(msg->param(0,true))) {
      static_cast<plaintab *>(tab)->action(msg->from()->nick(), msg->params(2, true));
      return;
    }
    // so we don't have a channel tab. Odd. Create one.
#ifdef IRCDEBUG
    debugtab->addRawLine("No tab for channel " + msg->param(0, true) + ", creating...");
#endif
    channeltab *tab = addNewChannelTab(msg->param(0, true));
    tab->action(msg->from()->nick(), msg->params(2, true));
  }
}


void ircsession::parseMode( ircmode *mode )
{
  // Deal only with modes on US. 
  // All other modes should be dealt with by the respective tabs
  
  if(mode->target().lower() == conn->me->nick().lower()) {
    servertab->mode(mode);
  }
}
